#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "mpi.h"

#define MSG_REPS 10
#define DEF_BUFFLEN 64
#define MTAG 99

#define BUFF_VAL(i) ((i)+1) /* expected value in ith element of buffer */

int main(int argc, char *argv[]) {
  int rank, nprocs;
  double *mbuf; int len, reps;
  int syncEndPoints, nflits, lenflit;
  int i, r;
  double t1, t2, tbar;
  int nErrs;

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &nprocs);

  len = (argc > 1)? atoi(argv[1]): DEF_BUFFLEN;
  reps = (len < 100? 100: (len < 1000? 10: 1)) * MSG_REPS;
  nflits = (argc > 2)? atoi(argv[2]): 1;
  if (nflits <= 0 || nflits > len) {
    if (rank == 0)
      printf("one-all ring: nflits=%d must be in the range 1..len=%d\n",
	     nflits, len);
    exit(1);
  }
  lenflit = (len + nflits - 1) / nflits; //last flit may be shorter
  syncEndPoints = (argc > 3)? atoi(argv[3]): 1;

  mbuf = calloc(len*nprocs, sizeof(double));
  if (rank == 0) // for checking, ensure different values on root
    for (i=0; i < len; i++)
      mbuf[i] = BUFF_VAL(i);

  MPI_Barrier(MPI_COMM_WORLD);
  MPI_Barrier(MPI_COMM_WORLD);
  t1 = MPI_Wtime();
  MPI_Barrier(MPI_COMM_WORLD);
  tbar = MPI_Wtime() - t1;

  if (rank==0)
    printf("one-all ring p=%d%s: %d words, %d flits, %d reps; t_bar=%.2es\n", 
	   nprocs, syncEndPoints? " sync e.p.": "", len, nflits, reps, tbar);

  t1 = 0.0;
  for (r=0; r < reps+1; r++) {
    if (r==1) t1 = MPI_Wtime(); //exclude rep 0 for warm-up
    for (i=0; i < nflits; i++) {
      int starti = 0, leni = lenflit; // only correct for nflits==1!
      if (rank > 0)
	MPI_Recv(&mbuf[starti], leni, MPI_DOUBLE, rank-1, MTAG, 
		 MPI_COMM_WORLD, MPI_STATUS_IGNORE);
      if (rank < nprocs-1)
	MPI_Send(&mbuf[starti], leni, MPI_DOUBLE, rank+1, MTAG, 
		 MPI_COMM_WORLD);
    }
    if (syncEndPoints) { // synchronize endpoint processes
      if (rank == nprocs-1)
	MPI_Send(mbuf, 0, MPI_DOUBLE, 0, MTAG, MPI_COMM_WORLD);
      if (rank == 0)
	MPI_Recv(mbuf, 0, MPI_DOUBLE, nprocs-1, MTAG, MPI_COMM_WORLD, 
		 MPI_STATUS_IGNORE);
    }
  } //for(r...)

  MPI_Barrier(MPI_COMM_WORLD);
  t2 = MPI_Wtime() - t1 - tbar;
  if (rank==0)
    printf("time: av t=%.2es bw=%.2eMB/s\n", t2/reps, 1.0E-6*len*reps/t2);

  nErrs = 0;
  for (i=0; i < len; i++)
    nErrs += (mbuf[i] != BUFF_VAL(i));
  if (nErrs > 0)
    printf("process %d: %d incorrect elements in buffer!\n", rank, nErrs);

  free(mbuf);
  MPI_Finalize();
  return 0;
} //main()
